/*
 * ext3rminator - a program to recover deleted data from an ext3 partition
 * Copyright (c) 2002-2003 Mike Hommey <mh@glandium.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <com_err.h>
#include <dirent.h>
#include "ext3rminator.h"

int read_count = 0;

char diskmap_signature[] = "EXT3RMIN";

int compare(const void *a, const void *b) {
  if (((__u32 *) a)[0] == ((__u32 *) b)[0])
    return 0;
  if (((__u32 *) a)[0] < ((__u32 *) b)[0])
    return -1;
  return 1;
}

/* Tell if a block looks like an indirect block or not.
 * Returns number of indirect pointers if indirect, -1 if not indirect, 0 if empty
 */
int is_indirect_block(ext3rminator *ext3rminator, __u32 *indirection)
{
  int i, first_null;
  /* Look every possible pointer in block */
  for (i = 0, first_null = -1; i < (1024 << ext3rminator->fs->super->s_log_block_size) / sizeof(__u32); i++) {
    /* Find the first NULL pointer */
    if (!(first_null > -1) && (indirection[i] == 0))
      first_null = i;
    /* If pointer is out of disk size, or if there are non-NULL pointers
     * AFTER the first NULL pointer, block is not indirect */
    if ((indirection[i] > ext3rminator->fs->super->s_blocks_count) ||
	((first_null > -1) && indirection[i])) {
      return -1;
    }
  }
  /* if first NULL pointer was not found, then set it to the number of pointers in a block */
  if (first_null == -1) first_null = (1024 << ext3rminator->fs->super->s_log_block_size) / sizeof(__u32);

  /* Now, we check that pointers are all different */
  if (first_null) {
    __u32 *tmp = malloc((1024 << ext3rminator->fs->super->s_log_block_size) * 2);
    int seq = 0;
    tmp[0] = indirection[0];
    for (i = 1; i < first_null; i++) {
      if (indirection[i] != indirection[i - 1] + 1) {
	tmp[++seq] = indirection[i - 1];
	tmp[++seq] = indirection[i];
      }
    }
    tmp[++seq] = indirection[i - 1];
    seq++;
    qsort(tmp, seq / 2, sizeof(__u32) * 2, compare);
    for (i = 1; i < seq / 2; i++) {
      if (tmp[i * 2] <= tmp[i * 2 - 1]) {
	free(tmp);
	return -1;
      }
    }
    free(tmp);
  }

  return first_null;
}

/* ext3rminator structure initialization */
ext3rminator *ext3r_create() {
  ext3rminator *result = NULL;

  /* Allocating memory for structure */
  result = malloc(sizeof(ext3rminator));
  if (! result) {
    fprintf(stderr,"ext3r_create(): %s.\n", strerror(errno));
    return NULL;
  }
  memset(result, 0, sizeof(ext3rminator));
  result->diskmap_fd = -1;

  return result;
}

/* Set device file name */
int ext3r_set_device(ext3rminator *ext3rminator, char *filename) {
  struct stat64 s;

  if (!ext3rminator || !filename) return 0;

  /* Verify file is regular or block device */
  if (stat64(filename, &s)) {
    fprintf(stderr, "Error while stating file %s: %s.\n", filename, strerror(errno));
    return 0;
  }
  if (! S_ISREG(s.st_mode) && ! S_ISBLK(s.st_mode)) {
    fprintf(stderr, "%s is neither a block device nor a regular file.\n", filename);
    return 0;
  }

  /* Copy filename */
  if (! (ext3rminator->device = malloc(strlen(filename) + 1))) {
    fprintf(stderr, "ext3r_set_device(): %s.\n", strerror(errno));
    return 0;
  }
  strncpy(ext3rminator->device, filename, strlen(filename) + 1);

  ext3rminator->device_is_file = S_ISREG(s.st_mode);

  return 1;
}

/* Set disk map file name */
int ext3r_set_diskmap(ext3rminator *ext3rminator, char *filename) {
  struct stat s;

  if (!ext3rminator || !filename) return 0;

  /* Check if the file exists ; do we need to read from it ? */
  if (stat(filename, &s)) {
    if (errno == ENOENT)
      ext3rminator->newdiskmap = 1;
  } else {
    if (!S_ISREG(s.st_mode)) {
      fprintf(stderr, "%s is not a regular file\n", filename);
      return 0;
    }
    if (s.st_size == 0) {
      fprintf(stderr, "Disk map file %s is empty\n", filename);
      return 0;
    }
    if (s.st_size < strlen(diskmap_signature) + sizeof(struct ext2_super_block)) {
      fprintf(stderr, "%s is too small to be a diskmap.\n", filename);
      return 0;
    }
  }  
  /* Copy filename */
  if (! (ext3rminator->diskmap = malloc(strlen(filename) + 1))) {
    fprintf(stderr, "ext3r_set_diskmap(): %s.\n", strerror(errno));
    return 0;
  }
  strncpy(ext3rminator->diskmap, filename, strlen(filename) + 1);

  return 1;
}

/* Set destination directory for extraction */
int ext3r_set_destdir(ext3rminator *ext3rminator, char *filename) {
  struct stat s;
  if (!ext3rminator || !filename) return 0;

  /* Check if the file exists ; do we need to read from it ? */
  if (stat(filename, &s)) {
    if (mkdir(filename, 0755) == -1) {
      fprintf(stderr, "Couldn't create %s directory: %s\n", filename, strerror(errno));
      return 0;
    }
  } else {
    if (!S_ISDIR(s.st_mode)) {
      fprintf(stderr, "%s is not a directory\n", filename);
      return 0;
    }
  }  
  /* Copy filename */
  if (! (ext3rminator->destdir = malloc(strlen(filename) + 1))) {
    fprintf(stderr, "ext3r_set_destdir(): %s.\n", strerror(errno));
    return 0;
  }
  strncpy(ext3rminator->destdir, filename, strlen(filename) + 1);

  return 1;
}

/* Initialize device */
int ext3r_init_device(ext3rminator *ext3rminator) {
  int retval;

  if (!ext3rminator || !ext3rminator->device) return 0;

  /* Opening device file */
  retval = ext2fs_open(ext3rminator->device, 0, 0, 0, unix_io_manager, &ext3rminator->fs);
  if (retval) {
    com_err(ext3rminator->device, retval, "while opening filesystem");
    goto error;
  }
  /* Initialize memory block map */
  ext3rminator->block_map = malloc(ext3rminator->fs->group_desc_count * ext3rminator->fs->blocksize * 8);
  if (! ext3rminator->block_map) {
    fprintf(stderr,"%s.\n", strerror(errno));
    goto error;
  }

  /* Print a little bit of information */
  printf("%s to explore: %s.\n", ext3rminator->device_is_file ? "File" : "Device", ext3rminator->device);
  printf(" %d inodes (%d free).\n" \
	 " %d blocks of %d bytes (%d free).\n" \
	 " Last mounted on %s\n",
	 ext3rminator->fs->super->s_inodes_count,
	 ext3rminator->fs->super->s_free_inodes_count,
	 ext3rminator->fs->super->s_blocks_count,
	 ext3rminator->fs->blocksize,
	 ext3rminator->fs->super->s_free_blocks_count,
	 ctime( (time_t *) &ext3rminator->fs->super->s_mtime));
  return 1;

  /* On error, free all allocated memory and close all opened files */
 error:
  if (ext3rminator->fs) {
    ext2fs_free(ext3rminator->fs);
    ext3rminator->fs = NULL;
  }
  return 0;
}

/* Initialize disk map */
int ext3r_init_diskmap(ext3rminator *ext3rminator) {
  int i;
  char signature_buffer[8];
  struct ext2_super_block superblock;

  if (!ext3rminator) return 0;

  if (ext3rminator->diskmap) {
    /* Open disk map file */
    if ((ext3rminator->diskmap_fd = open(ext3rminator->diskmap, O_RDWR | O_CREAT, 00666)) == -1) {
      fprintf(stderr, "Error while opening file %s: %s.\n", ext3rminator->diskmap, strerror(errno));
      goto error;
    }

    printf("Using %sfile %s to store disk map.\n", ext3rminator->newdiskmap ? "" : "existing ", ext3rminator->diskmap);
  } else {
    printf("Disk map won't be stored.\n");
  }
  /* Get data from file if it existed, from device otherwise */
  if (!ext3rminator->newdiskmap && ext3rminator->diskmap) {
    /* Read file signature */
    if (read(ext3rminator->diskmap_fd, signature_buffer, 8) == 8) {
      if (strncmp(signature_buffer, diskmap_signature, 8)) {
	fprintf(stderr, "File %s is not an ext3rminator disk map file.\n", ext3rminator->diskmap);
	goto error;
      }
    } else {
      fprintf(stderr, "Error while reading file %s: %s.\n", ext3rminator->diskmap, errno ? strerror(errno) : "Premature end of file");
      goto error;
    }
    /* Read superblock from disk map file */
    if (read(ext3rminator->diskmap_fd, &superblock, sizeof(struct ext2_super_block)) != sizeof(struct ext2_super_block)) {
      fprintf(stderr, "Error while reading superblock from file %s: %s\n", ext3rminator->diskmap, errno ? strerror(errno) : "Premature end of file");
      goto error;
    }
    /* Compare superblock read from file with superblock read from device */
    if (strncmp((char *)ext3rminator->fs->super, (char *)&superblock, sizeof(struct ext2_super_block))) {
      fprintf(stderr, "Superblock mismatch.\n");
      goto error;
    }
    /* Read block map from file */
    if (read(ext3rminator->diskmap_fd, ext3rminator->block_map, ext3rminator->fs->group_desc_count * ext3rminator->fs->blocksize * 8) != ext3rminator->fs->group_desc_count * ext3rminator->fs->blocksize * 8) {
      fprintf(stderr, "Error while reading block map from file %s: %s\n", ext3rminator->diskmap, errno ? strerror(errno) : "Premature end of file");
      goto error;
    }
    /* Counting free blocks */
    for (i = 0; i < ext3rminator->fs->group_desc_count * ext3rminator->fs->blocksize * 8; i++) {
      if (!BLOCK_IS_SET(ext3rminator->block_map, i, BLOCK_USED))
	ext3rminator->free_blocks++;
    }
  } else {
    int group;
    char *buffer;

    /* Reset memory block map */
    memset(ext3rminator->block_map, 0, ext3rminator->fs->group_desc_count * ext3rminator->fs->blocksize * 8);

    /* Allocate memory for block map buffer */
    buffer = malloc(ext3rminator->fs->blocksize);
    if (! buffer) {
      fprintf(stderr,"%s.\n", strerror(errno));
      goto error;
    }

    /* Read group block bitmaps */
    for(group = 0; group < ext3rminator->fs->group_desc_count; group++) {
      printf(" Reading block group bitmap %d of %d.\r", group + 1, ext3rminator->fs->group_desc_count);
      fflush(stdout);
      if (io_channel_read_blk(ext3rminator->fs->io, ext3rminator->fs->group_desc[group].bg_block_bitmap, 1, buffer) == -1) {
	fprintf(stderr, "Error while reading block bitmaps\n");
	goto error;
      }
      for (i = 0; i < ext3rminator->fs->blocksize << 3; i++) {
	int b;
	if (ext3rminator->fs->flags & EXT2_FLAG_SWAP_BYTES)
	  b = ext2fs_swab32(buffer[i >> 3]);
	else
	  b = buffer[i >> 3];
	if (b & (1 << (i & 0x07))) {
	  BLOCK_SET_EXACT(ext3rminator->block_map, ext3rminator->fs->super->s_blocks_per_group * group + i, BLOCK_USED);
	} else
	  ext3rminator->free_blocks++;
      }
    }
    free(buffer);
    printf(" Read %d block group bitmaps.                   \n", group);
  }
  printf(" %d blocks to explore.\n\n", ext3rminator->free_blocks);
  return 1;

 error:
  if (ext3rminator->diskmap_fd != -1) {
    close(ext3rminator->diskmap_fd);
    ext3rminator->diskmap_fd = -1;
  }
  if (ext3rminator->block_map) {
    free(ext3rminator->block_map);
    ext3rminator->block_map = 0;
  }
  return 0;
}

/* Explore file system to fill disk map */
int ext3r_explore(ext3rminator *ext3rminator) {
  __u32 *buffer = NULL;
  int block, i, j, size = 0, indirect = 0, valid_indirect = 0, empty = 0, free_block = 0, pointed = 0, multiple_pointed = 0, dble;

  if (ext3rminator->newdiskmap || ! ext3rminator->diskmap) {
    printf("Exploring file system to fill disk map.\n");

    /* Allocate memory for block buffer */
    buffer = malloc(ext3rminator->fs->blocksize);
    if (! buffer) {
      fprintf(stderr,"%s.\n", strerror(errno));
      return 0;
    }
    
    for(block = 0; block < ext3rminator->fs->super->s_blocks_count; block++) {
      /* Lookup every block marked as free in block map */
      if (!BLOCK_IS_SET(ext3rminator->block_map, block, BLOCK_USED)) {
	if (((++free_block) % 1000 == 0) || (size > 0)) {
	  printf(" Reading free block %d of %d. Found %d empty and %d indirect blocks\r", free_block, ext3rminator->free_blocks, empty, indirect);
	  fflush(stdout);
	}
	/* Read a block */
	if (io_channel_read_blk(ext3rminator->fs->io, block, 1, buffer) == -1) {
	  goto error;
	}
	/* Swap bits if needed */
	if (ext3rminator->fs->flags & EXT2_FLAG_SWAP_BYTES) {
	  int i;
	  for (i=0; i < ext3rminator->fs->blocksize >> 2; i++)
	    ((__u32 *)buffer)[i] = ext2fs_swab32(((__u32 *)buffer)[i]);
	}

	/* Mark as indirect if it looks like */
	if ((size = is_indirect_block(ext3rminator, buffer)) > 0) {
	  BLOCK_SET(ext3rminator->block_map, block, BLOCK_VALID | BLOCK_INDIRECT);
	  indirect++;
	  /* Mark all pointed blocks and make a first check of indirect block validity */
	  for (i = 0; i < size; i++) {
	    /* If pointed block is used, block is invalid indirect block */
	    if (BLOCK_IS(ext3rminator->block_map, buffer[i], BLOCK_USED))
	      BLOCK_UNSET(ext3rminator->block_map, block, BLOCK_VALID);
	  }
	  if (BLOCK_IS_SET(ext3rminator->block_map, block, BLOCK_VALID))
	    for (i = 0; i < size; i++) {
	      /* If pointed block was already pointed, set it multiple */
	      if (BLOCK_IS(ext3rminator->block_map, buffer[i], BLOCK_POINTED))
		BLOCK_SET(ext3rminator->block_map, buffer[i], BLOCK_MULTIPLE_POINTED);
	      else if (BLOCK_IS_FREE(ext3rminator->block_map, buffer[i]))
		BLOCK_SET(ext3rminator->block_map, buffer[i], BLOCK_POINTED);
	    }
	} else if (size == 0) {
	  /* Mark as empty if it is */
	  BLOCK_SET_EXACT(ext3rminator->block_map, block, BLOCK_EMPTY);
	  empty++;
	}
      }
    }
    printf(" Read %d free blocks.                                                    \n Now looking for double indirect blocks.\n", free_block);
    for (i = 0; i < ext3rminator->fs->super->s_blocks_count; i++) {
      if (BLOCK_IS(ext3rminator->block_map, i, BLOCK_VALID | BLOCK_INDIRECT)) {
	/* Read a block */
	if (io_channel_read_blk(ext3rminator->fs->io, i, 1, buffer) == -1) {
	  goto error;
	}
	/* Swap bits if needed */
	if (ext3rminator->fs->flags & EXT2_FLAG_SWAP_BYTES) {
	  for (j = 0; j < ext3rminator->fs->blocksize >> 2; j++)
	    buffer[j] = ext2fs_swab32(buffer[j]);
	}
	dble = 1;
	for (j = 0; j < ext3rminator->fs->blocksize >> 2; j++)
	  if (! buffer[j])
	    break;
	  else if (! BLOCK_IS(ext3rminator->block_map, buffer[j], BLOCK_VALID | BLOCK_INDIRECT | BLOCK_POINTED))
	    dble = 0;
	if (dble)
	  /* This is naughty ; should really be cleaner ; BTW, we don't support triple indirections */
	  BLOCK_UNSET(ext3rminator->block_map, i, BLOCK_TRIPLE_INDIRECT);
      }
    }
  } else {
    printf("Using disk map from previous disk exploration.\n");
  }

  indirect = 0;
  empty = 0;
  /* Counting free blocks */
  for (i = 0; i < ext3rminator->fs->super->s_blocks_count; i++) {
    if (BLOCK_IS_SET(ext3rminator->block_map, i, BLOCK_VALID) &&
	BLOCK_IS_SET(ext3rminator->block_map, i, BLOCK_INDIRECT))
      valid_indirect++;
    if (BLOCK_IS_SET(ext3rminator->block_map, i, BLOCK_INDIRECT))
      indirect++;
    if (BLOCK_IS(ext3rminator->block_map, i, BLOCK_EMPTY))
      empty++;
    if (BLOCK_IS_SET(ext3rminator->block_map, i, BLOCK_MULTIPLE_POINTED))
      multiple_pointed++;
    if (BLOCK_IS_SET(ext3rminator->block_map, i, BLOCK_POINTED))
      pointed++;
  }
  printf(" %d empty blocks.\n" \
	 " %d valid indirect blocks (%d total indirect blocks)\n" \
	 " %d pointed blocks (%d multiple pointed blocks)\n\n",
	 empty,
	 valid_indirect, indirect,
	 pointed, multiple_pointed);

  if (buffer)
    free(buffer);

  if (ext3rminator->diskmap && !ext3r_save_diskmap(ext3rminator))
    goto error;

  return 1;

 error:
  free(buffer);
  return 0;
}

/* virtual inode */
struct ext2_inode the_inode;

/* Function to return the current read virtual inode to libext2fs functions */
errcode_t read_inode(ext2_filsys fs, ext2_ino_t ino, struct ext2_inode *inode) {
  memcpy(inode, &the_inode, sizeof(struct ext2_inode));
  return 0;
}

/* Extracts data from filesystem */
/* OMFG what a mess ! */
int ext3r_extract(ext3rminator *ext3rminator) {
  __u32 block = 0;
  __u32 *buffer;
  int i, fd, p;
  ext2_file_t file;
  char *filename;
  unsigned int count = 0;
  unsigned long long byte_count = 0;

  if (ext3rminator->destdir) {
    DIR *dir;
    struct dirent *file;
    unsigned int tmp;

    printf("Data will be extracted in directory %s.\n", ext3rminator->destdir);
    dir = opendir(ext3rminator->destdir);
    while ((file = readdir(dir))) {
      if (file->d_type == 8) { // WHERE CAN I GET A #DEFINE IN /USR/INCLUDE FOR THE FILE D_TYPE, BORDEL DE MERDE ????
	if (sscanf(file->d_name, "%08x", &tmp)) {
	  if (tmp > block)
	    block = (__u32) tmp;
	} else {
	  block = 0;
	  break;
	}
      }
    }
    closedir(dir);
    if (block) {
      printf("It seems that a previous extraction has been done to this directory,\nTrying to continue it...\n");
    } else {
      block = ext3rminator->fs->super->s_first_data_block + EXT2_NDIR_BLOCKS;
    }
  } else {
    printf("Data won't be extracted.\n");
    return 1;
  }
  buffer = malloc(ext3rminator->fs->blocksize);
  ext3rminator->fs->read_inode = read_inode;
  filename = malloc(strlen(ext3rminator->destdir) + 10);

  memset(&the_inode, 0, sizeof(the_inode));
  for(; block < ext3rminator->fs->super->s_blocks_count; block++) {
    if (BLOCK_IS(ext3rminator->block_map, block, BLOCK_VALID | BLOCK_INDIRECT)) {
      for (i = 0; i < EXT2_NDIR_BLOCKS; i++)
	the_inode.i_block[i] = block - EXT2_NDIR_BLOCKS + i;
      the_inode.i_block[EXT2_IND_BLOCK] = block;

      /* Read a block */
      if (io_channel_read_blk(ext3rminator->fs->io, block, 1, buffer) == -1) {
	goto error;
      }
      /* Swap bits if needed */
      if (ext3rminator->fs->flags & EXT2_FLAG_SWAP_BYTES) {
	for (i = 0; i < ext3rminator->fs->blocksize >> 2; i++)
	  buffer[i] = ext2fs_swab32(buffer[i]);
      }
      for (i = 0; i < ext3rminator->fs->blocksize >> 2; i++)
	if (buffer[i] == 0)
	  break;
      p = buffer[i - 1] + 1;
      the_inode.i_size = (EXT2_NDIR_BLOCKS + i - 1) * ext3rminator->fs->blocksize;
      if (io_channel_read_blk(ext3rminator->fs->io, buffer[i - 1], 1, buffer) == -1) {
	goto error;
      }
      for (i = ext3rminator->fs->blocksize; i > 0; i--)
	if (((char *)buffer)[i - 1] != 0)
	  break;

      the_inode.i_size += i;
      if ((the_inode.i_size == (EXT2_NDIR_BLOCKS + (ext3rminator->fs->blocksize >> 2)) * ext3rminator->fs->blocksize) && (BLOCK_IS(ext3rminator->block_map, p, BLOCK_VALID | BLOCK_DOUBLE_INDIRECT))) {
	the_inode.i_block[EXT2_DIND_BLOCK] = p;
	if (io_channel_read_blk(ext3rminator->fs->io, p, 1, buffer) == -1) {
	  goto error;
	}
	if (ext3rminator->fs->flags & EXT2_FLAG_SWAP_BYTES) {
	  for (i = 0; i < ext3rminator->fs->blocksize >> 2; i++)
	    buffer[i] = ext2fs_swab32(buffer[i]);
	}
	for (i = 0; i < ext3rminator->fs->blocksize >> 2; i++)
	  if (buffer[i] == 0)
	    break;
	the_inode.i_size += (i - 1) * ext3rminator->fs->blocksize * ext3rminator->fs->blocksize;
	if (io_channel_read_blk(ext3rminator->fs->io, buffer[i - 1], 1, buffer) == -1) {
	  goto error;
	}
	if (ext3rminator->fs->flags & EXT2_FLAG_SWAP_BYTES) {
	  for (i = 0; i < ext3rminator->fs->blocksize >> 2; i++)
	    buffer[i] = ext2fs_swab32(buffer[i]);
	}
	for (i = 0; i < ext3rminator->fs->blocksize >> 2; i++)
	  if (buffer[i] == 0)
	    break;
	the_inode.i_size += (i - 1) * ext3rminator->fs->blocksize;
	if (io_channel_read_blk(ext3rminator->fs->io, buffer[i - 1], 1, buffer) == -1) {
	  goto error;
	}
	for (i = ext3rminator->fs->blocksize; i > 0; i--)
	  if (((char *)buffer)[i - 1] != 0)
	    break;
	
	the_inode.i_size += i;
      }
      sprintf(filename, "%s/%08x", ext3rminator->destdir, block - EXT2_NDIR_BLOCKS);
      if ((fd = open(filename, O_CREAT | O_WRONLY | O_TRUNC, 0644)) == -1) {
	fprintf(stderr, "Error while opening %s output file: %s\n", filename, strerror(errno));
	goto error;
      }
      if (ext2fs_file_open(ext3rminator->fs, 0, 0, &file) == 0) {
	int got;
	while(! ext2fs_file_read(file, (char *)buffer, ext3rminator->fs->blocksize, &got) && got) {
	  if (write(fd, (char *) buffer, got) != got) {
	    fprintf(stderr,"Error while writing %s output file: %s\n", filename, strerror(errno));
	    goto error;
	  }
	}
	ext2fs_file_close(file);
	close(fd);
	printf(" Extracted %llu bytes in %u files\r", byte_count += the_inode.i_size , ++count);
	fflush(stdout);
      } else {
	fprintf(stderr, "Error while opening file at block %08x\n", block - EXT2_NDIR_BLOCKS);
      }
    }
  }
  printf("\n");

  return 1;
 error:
  free(buffer);
  free(filename);
  return 0;
}

/* Save disk map */
int ext3r_save_diskmap(ext3rminator *ext3rminator) {

  if (!ext3rminator || !ext3rminator->diskmap) return 0;

  if (lseek(ext3rminator->diskmap_fd, 0, SEEK_SET) == -1) {
    fprintf(stderr, "Error seeking at beginning of file %s: %s.\n", ext3rminator->diskmap, strerror(errno));
    goto error;
  }
  /* Write signature to file */
  if (write(ext3rminator->diskmap_fd, diskmap_signature, 8) != 8) {
    fprintf(stderr, "Error writing to file %s: %s.\n", ext3rminator->diskmap, strerror(errno));
    goto error;
  }
  /* Write a copy of superblock to file */
  if (write(ext3rminator->diskmap_fd, ext3rminator->fs->super, sizeof(struct ext2_super_block)) != sizeof(struct ext2_super_block)) {
    fprintf(stderr, "Error writing to file %s: %s.\n", ext3rminator->diskmap, strerror(errno));
    goto error;
  }
  /* Write initial block map to file */
  if (write(ext3rminator->diskmap_fd, ext3rminator->block_map, ext3rminator->fs->group_desc_count * ext3rminator->fs->blocksize * 8) != ext3rminator->fs->group_desc_count * ext3rminator->fs->blocksize * 8) {
    fprintf(stderr, "Error writing to file %s: %s.\n", ext3rminator->diskmap, strerror(errno));
    goto error;
  }
  return 1;
 error:
  return 0;
}

int ext3r_destroy(ext3rminator *ext3rminator) {
  if (! ext3rminator)
    return 0;
  if (ext3rminator->fs)
    ext2fs_free(ext3rminator->fs);
  if (ext3rminator->diskmap_fd)
    close(ext3rminator->diskmap_fd);
  if (ext3rminator->diskmap)
    free(ext3rminator->diskmap);
  if (ext3rminator->device)
    free(ext3rminator->device);
  if (ext3rminator->block_map)
    free(ext3rminator->block_map);
  free(ext3rminator);

  return 1;
}
